import os
from unittest import TestCase
from unittest.mock import patch

from shub_workflow.crawl import CrawlManager, PeriodicCrawlManager, GeneratorCrawlManager
from shub_workflow.utils.contexts import script_args
from shub_workflow.script import SpiderName, Outcome


class TestManager(CrawlManager):

    name = "test"


class PeriodicTestManager(PeriodicCrawlManager):

    name = "test"


class ListTestManager(GeneratorCrawlManager):

    name = "test"
    default_max_jobs = 2

    def set_parameters_gen(self):
        parameters_list = [
            {"argA": "valA"},
            {"argA": "valB"},
            {"argB": "valC"},
        ]
        for args in parameters_list:
            yield args

    def bad_outcome_hook(self, spider, outcome, job_args_override, jobkey):
        spider_args = job_args_override.setdefault("spider_args", {})
        if "argR" not in spider_args:
            spider_args.update({"argR": "valR"})
            self.add_job(spider, job_args_override)
        elif "argS" not in spider_args:
            spider_args.update({"argS": "valS"})
            self.add_job(spider, job_args_override)


class TestManagerWithSpider(CrawlManager):

    name = "test"
    spider = SpiderName("myimplicitspider")


@patch("shub_workflow.script.BaseScript.get_jobs")
@patch("shub_workflow.script.BaseScript.add_job_tags")
class CrawlManagerTest(TestCase):
    def setUp(self):
        os.environ["SH_APIKEY"] = "ffff"
        os.environ["PROJECT_ID"] = "999"

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):

        with script_args(["myspider"]):
            manager = TestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1"]
        manager._on_start()

        # first loop: schedule spider
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call("myspider", units=None, job_settings={})

        # second loop: spider still running. Continue.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)

        # third loop: spider is finished. Stop.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        mocked_super_schedule_spider.reset_mock()
        result = next(manager._run_loops())

        self.assertFalse(result)
        self.assertFalse(mocked_super_schedule_spider.called)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_implicit_spider(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):

        with script_args([]):
            manager = TestManagerWithSpider()

        mocked_super_schedule_spider.side_effect = ["999/1/1"]
        manager._on_start()

        # first loop: schedule spider
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call("myimplicitspider", units=None, job_settings={})

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_bad_outcome(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):

        with script_args(["myspider"]):
            manager = TestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1"]
        manager._on_start()

        # first loop: schedule spider
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call("myspider", units=None, job_settings={})

        # second loop: spider still running. Continue.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)

        # third loop: spider is cancelled. Stop. Manager must be closed with cancelled close reason
        manager.is_finished = lambda x: Outcome("cancelled") if x == "999/1/1" else None
        mocked_super_schedule_spider.reset_mock()
        result = next(manager._run_loops())

        self.assertFalse(result)
        self.assertFalse(mocked_super_schedule_spider.called)

        self.assertEqual(manager.get_close_reason(), "cancelled")

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_with_resume(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
        with script_args(["myspider", "--flow-id=3a20"]):
            manager = TestManager()

        mocked_get_jobs_side_effect = [
            # the resumed job
            [{"tags": ["FLOW_ID=3a20", "NAME=test", "OTHER=other"], "key": "999/10/1"}],
            # the owned running job
            [{"spider": "myspider", "key": "999/1/1", "tags": ["FLOW_ID=3a20", "PARENT_NAME=test"]}],
            # the owned finished jobs
            [],
        ]
        mocked_get_jobs.side_effect = mocked_get_jobs_side_effect
        manager._on_start()
        self.assertTrue(manager.is_resumed)
        self.assertEqual(len(manager._running_job_keys), 1)

        for v in manager._running_job_keys.values():
            self.assertEqual(set(v[1].keys()), {"spider_args", "tags"})

        self.assertEqual(mocked_get_jobs.call_count, len(mocked_get_jobs_side_effect))
        mocked_add_job_tags.assert_any_call(tags=["FLOW_ID=3a20", "NAME=test", "OTHER=other"])

        # first loop: spider still running in workflow. Continue.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)

        # second loop: spider is finished. Stop.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        result = next(manager._run_loops())

        self.assertFalse(result)
        self.assertFalse(mocked_super_schedule_spider.called)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_with_resume_not_found(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        with script_args(["myspider", "--flow-id=3a20"]):
            manager = TestManager()

        mocked_get_jobs_side_effect = [
            # the not resumed job (different flow id)
            [{"tags": ["FLOW_ID=3344", "NAME=othertest"], "key": "999/10/1"}],
        ]
        mocked_get_jobs.side_effect = mocked_get_jobs_side_effect
        manager._on_start()
        self.assertFalse(manager.is_resumed)
        self.assertEqual(len(manager._running_job_keys), 0)
        self.assertEqual(mocked_get_jobs.call_count, len(mocked_get_jobs_side_effect))

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_with_resume_not_owned(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        with script_args(["myspider", "--flow-id=3a20"]):
            manager = TestManager()

        mocked_get_jobs.side_effect = [
            # the resumed job
            [{"tags": ["FLOW_ID=3a20", "NAME=test"], "key": "999/10/1"}],
            # the not owned job
            [{"key": "999/1/1", "tags": ["FLOW_ID=3a20", "PARENT_NAME=testa"]}],
            # owned finished jobs
            [],
        ]
        manager._on_start()
        self.assertTrue(manager.is_resumed)
        self.assertEqual(len(manager._running_job_keys), 0)

        # first loop: no spider, schedule one.
        manager.is_finished = lambda x: None
        mocked_super_schedule_spider.side_effect = ["999/1/2"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)

        # second loop: spider is finished. Stop.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/2" else None
        result = next(manager._run_loops())

        self.assertFalse(result)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_periodic(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
        with script_args(["myspider"]):
            manager = PeriodicTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1"]
        manager._on_start()

        # first loop: schedule spider
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call("myspider", units=None, job_settings={})

        # second loop: spider still running. Continue.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)

        # third loop: spider is finished. Schedule again.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        mocked_super_schedule_spider.reset_mock()
        mocked_super_schedule_spider.side_effect = ["999/1/2"]
        result = next(manager._run_loops())

        self.assertTrue(result)
        mocked_super_schedule_spider.assert_any_call("myspider", units=None, job_settings={})

        # four loop: spider is cancelled. Schedule again.
        manager.is_finished = lambda x: Outcome("cancelled") if x == "999/1/2" else None
        mocked_super_schedule_spider.reset_mock()
        mocked_super_schedule_spider.side_effect = ["999/1/3"]
        result = next(manager._run_loops())

        self.assertTrue(result)
        mocked_super_schedule_spider.assert_any_call("myspider", units=None, job_settings={})

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_list_bad_outcome_hook(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        with script_args(["myspider"]):
            manager = ListTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/1/2", "999/1/3", "999/1/4", "999/1/5"]
        manager._on_start()

        # first loop: schedule spider with first set of arguments
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["JOBSEQ=0000000001"], job_settings={}
        )
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valB", tags=["JOBSEQ=0000000002"], job_settings={}
        )

        # second loop: still no job finished. Wait for a free slot
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)

        # third loop: finish one job. We can schedule last one with third set of arguments
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argB="valC", tags=["JOBSEQ=0000000003"], job_settings={}
        )

        # fourth loop: waiting jobs to finish
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)

        # fifth loop: second job finished with failed outcome. Retry according
        # to bad outcome hook
        manager.is_finished = lambda x: Outcome("cancelled (stalled)") if x == "999/1/2" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valB", tags=["JOBSEQ=0000000002.r1"], argR="valR", job_settings={}
        )

        # sixth loop: third job finished.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/3" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)

        # seventh loop: retried job failed again. Retry with new argument.
        manager.is_finished = lambda x: Outcome("memusage_exceeded") if x == "999/1/4" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 5)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valB", tags=["JOBSEQ=0000000002.r2"], argR="valR", job_settings={}
        )

        # eighth loop: retried job finished. Exit.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/5" else None
        result = next(manager._run_loops())
        self.assertFalse(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 5)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_list_explicit_spider(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            default_max_jobs = 2
            spider = SpiderName("myspider")

            def set_parameters_gen(self):
                parameters_list = [
                    {"argA": "valA"},
                    {"argA": "valB", "spider": "myspidertwo"},
                ]
                for args in parameters_list:
                    yield args

        with script_args([]):
            manager = _ListTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/1/2"]
        manager._on_start()

        # first loop: schedule spider with first set of arguments
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["JOBSEQ=0000000001"], job_settings={}
        )
        mocked_super_schedule_spider.assert_any_call(
            "myspidertwo", units=None, argA="valB", tags=["JOBSEQ=0000000002"], job_settings={}
        )

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_list_scrapy_cloud_params(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            default_max_jobs = 2
            spider = SpiderName("myspider")

            def set_parameters_gen(self):
                parameters_list = [
                    {
                        "argA": "valA",
                        "units": 2,
                        "tags": ["CHECKED"],
                        "project_id": 999,
                        "job_settings": {"CONCURRENT_REQUESTS": 2},
                    },
                ]
                for args in parameters_list:
                    yield args

        with script_args([]):
            manager = _ListTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/1/2"]
        manager._on_start()

        # first loop: schedule spider with first set of arguments
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "myspider",
            units=2,
            argA="valA",
            job_settings={"CONCURRENT_REQUESTS": 2},
            project_id=999,
            tags=["CHECKED", "JOBSEQ=0000000001"],
        )
        jobid = GeneratorCrawlManager.get_job_unique_id(
            {"spider": SpiderName("myspider"), "spider_args": {"argA": "valA"}}
        )
        self.assertTrue(jobid in manager._jobuids)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_list_with_resume(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            default_max_jobs = 2
            spider = SpiderName("myspider")

            def set_parameters_gen(self):
                parameters_list = [
                    {"argA": "valA"},
                    {"argA": "valB", "spider": "myspidertwo"},
                ]
                for args in parameters_list:
                    yield args

        with script_args(["--flow-id=3a20"]):
            manager = _ListTestManager()

        mocked_get_jobs.side_effect = [
            # the resumed job
            [{"tags": ["FLOW_ID=3a20", "NAME=test"], "key": "999/10/1"}],
            # running spiders
            [],
            # finished spiders
            [
                {
                    "spider": "myspider",
                    "key": "999/1/1",
                    "tags": ["FLOW_ID=3a20", "PARENT_NAME=test", "JOBSEQ=0000000001"],
                    "spider_args": {"argA": "valA"},
                }
            ],
        ]
        mocked_super_schedule_spider.side_effect = ["999/2/1"]
        manager._on_start()
        self.assertTrue(manager.is_resumed)
        self.assertEqual(len(manager._running_job_keys), 0)

        # first loop: only second task will be scheduled. First one already completed before resuming.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "myspidertwo", argA="valB", tags=["JOBSEQ=0000000002"], job_settings={}, units=None
        )

        # second loop: finished second spider. Finish execution
        manager.is_finished = lambda x: Outcome("finished") if x == "999/2/1" else None
        result = next(manager._run_loops())
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        self.assertFalse(result)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_schedule_spider_list_resumed_running_job_with_bad_outcome(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        class _ListTestManager(ListTestManager):

            name = "test"
            default_max_jobs = 2
            spider = SpiderName("myspider")

            def set_parameters_gen(self):
                parameters_list = [
                    {"argA": "valA"},
                ]
                for args in parameters_list:
                    yield args

        with script_args(["--flow-id=3a20"]):
            manager = _ListTestManager()

        mocked_get_jobs.side_effect = [
            # the resumed job
            [{"tags": ["FLOW_ID=3a20", "NAME=test"], "key": "999/10/1"}],
            # running spiders
            [
                {
                    "spider": "myspider",
                    "key": "999/1/1",
                    "tags": ["FLOW_ID=3a20", "PARENT_NAME=test", "JOBSEQ=0000000001"],
                    "spider_args": {"argA": "valA"},
                }
            ],
            # finished spiders
            [],
        ]
        mocked_super_schedule_spider.side_effect = ["999/2/1"]
        manager._on_start()
        self.assertTrue(manager.is_resumed)
        self.assertEqual(len(manager._running_job_keys), 1)

        for v in manager._running_job_keys.values():
            self.assertEqual(set(v[1].keys()), {"spider_args", "tags"})

        # first loop: acquire running job.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 0)

        # second loop: second job finished with failed outcome. Retry according
        # to bad outcome hook
        manager.is_finished = lambda x: Outcome("cancelled (stalled)")
        mocked_super_schedule_spider.side_effect = ["999/2/2"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "myspider",
            units=None,
            argA="valA",
            tags=["FLOW_ID=3a20", "PARENT_NAME=test", "JOBSEQ=0000000001.r1"],
            argR="valR",
            job_settings={},
        )

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_default_bad_outcome_no_retry(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            default_max_jobs = 1
            spider = SpiderName("myspider")

            def set_parameters_gen(self):
                parameters_list = [
                    {"argA": "valA"},
                ]
                for args in parameters_list:
                    yield args

        with script_args([]):
            manager = _ListTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1"]
        manager._on_start()

        # first loop: schedule spider
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["JOBSEQ=0000000001"], job_settings={}
        )

        # second loop: finish job with bad outcome, but there is no retry. Stop.
        manager.is_finished = lambda x: Outcome("cancelled (stalled)") if x == "999/1/1" else None
        result = next(manager._run_loops())
        self.assertFalse(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_default_bad_outcome_with_retries(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):
        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            default_max_jobs = 2
            spider = SpiderName("myspider")

            MAX_RETRIES = 2

            def set_parameters_gen(self):
                parameters_list = [
                    {"argA": "valA"},
                    {"argB": "valB"},
                ]
                for args in parameters_list:
                    yield args

        with script_args([]):
            manager = _ListTestManager()

        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/2/1"]
        manager._on_start()

        # first loop: schedule spiders
        result = next(manager._run_loops())

        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["JOBSEQ=0000000001"], job_settings={}
        )

        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argB="valB", tags=["JOBSEQ=0000000002"], job_settings={}
        )

        # second loop: finish first job with bad outcome, retry 1.
        manager.is_finished = lambda x: Outcome("cancelled (stalled)") if x == "999/1/1" else None
        mocked_super_schedule_spider.side_effect = ["999/1/2"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["RETRIED_FROM=999/1/1", "JOBSEQ=0000000001.r1"], job_settings={}
        )

        # second loop: second job finishes with "cancelled", don't retry it.
        manager.is_finished = lambda x: Outcome("cancelled") if x == "999/2/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)

        # third loop: first job finishes again with abnormal reason, retry it.
        manager.is_finished = lambda x: Outcome("cancelled (stalled)") if x == "999/1/2" else None
        mocked_super_schedule_spider.side_effect = ["999/1/3"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        mocked_super_schedule_spider.assert_any_call(
            "myspider", units=None, argA="valA", tags=["RETRIED_FROM=999/1/2", "JOBSEQ=0000000001.r2"], job_settings={}
        )

        # fourth loop: first job finishes again with abnormal reason, but max retries reached. Stop.
        manager.is_finished = lambda x: Outcome("cancelled (stalled)") if x == "999/1/3" else None
        result = next(manager._run_loops())
        self.assertFalse(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)

        manager._close()

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_max_jobs_per_spider(self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs):

        """Notice delayed jobs queue got filled in first loop with all next jobs params.
        This is unavoidable if we don't have information about the spiders that
        will come in new next params. If crawl manager stops prematurely, all the delayed
        params will be lost. But effect can be limited. See next test."""

        spider_list = ("spiderA", "spiderB", "spiderC")

        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            max_jobs_per_spider = 1
            spider = SpiderName("")

            def set_parameters_gen(self):
                for i in range(100):
                    for spider in spider_list:
                        yield {"spider": spider, "argA": i}

        with script_args([]):
            manager = _ListTestManager()

        manager._on_start()

        # first loop.

        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/2/1", "999/3/1"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)

        for seq, spider in enumerate(spider_list, start=1):
            mocked_super_schedule_spider.assert_any_call(
                spider, units=None, argA=0, tags=[f"JOBSEQ={seq:010d}"], job_settings={}
            )
        self.assertEqual(len(manager.get_delayed_jobs()), 297)

        # second loop. No job finished.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)

        # third loop. spiderB job finished. Can schedule new spiderB job.
        mocked_super_schedule_spider.side_effect = ["999/2/2"]
        manager.is_finished = lambda x: Outcome("finished") if x == "999/2/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        mocked_super_schedule_spider.assert_any_call(
            "spiderB", units=None, argA=1, tags=["JOBSEQ=0000000004"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 296)

        # fourth loop. spiderB job finished. Can schedule new spiderB job.
        mocked_super_schedule_spider.side_effect = ["999/2/3"]
        manager.is_finished = lambda x: Outcome("finished") if x == "999/2/2" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 5)
        mocked_super_schedule_spider.assert_any_call(
            "spiderB", units=None, argA=2, tags=["JOBSEQ=0000000005"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 295)

        # fifth loop. spiderC job finished. Can schedule new spiderC job.
        mocked_super_schedule_spider.side_effect = ["999/3/2"]
        manager.is_finished = lambda x: Outcome("finished") if x == "999/3/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 6)
        mocked_super_schedule_spider.assert_any_call(
            "spiderC", units=None, argA=1, tags=["JOBSEQ=0000000006"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 294)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_max_jobs_per_spider_with_max_next_params_limit(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        """Delayed jobs queue got filled in 5th loop with all next jobs params, when no new
        param is yielded. This is unavoidable if we don't have information about the spiders
        that will come in new next params. If crawl manager stops prematurely, all the delayed
        params will be lost. But effect can be limited further. See next test."""

        spider_list = ("spiderA", "spiderB", "spiderC")

        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            max_jobs_per_spider = 1
            spider = SpiderName("")

            def set_parameters_gen(self):
                for i in range(100):
                    for spider in spider_list:
                        yield {"spider": spider, "argA": i}

            def get_max_next_params(self) -> int:
                return min(1, super().get_max_next_params())

        with script_args([]):
            manager = _ListTestManager()

        manager._on_start()

        # 1st loop. Schedule spiderA
        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/2/1", "999/3/1"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=0, tags=["JOBSEQ=0000000001"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 2nd loop. Schedule spiderB
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)
        mocked_super_schedule_spider.assert_any_call(
            "spiderB", units=None, argA=0, tags=["JOBSEQ=0000000002"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 3rd loop. SpiderA finishes, but spiderC is next.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)
        mocked_super_schedule_spider.assert_any_call(
            "spiderC", units=None, argA=0, tags=["JOBSEQ=0000000003"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 4th loop. Schedule spiderA
        mocked_super_schedule_spider.side_effect = ["999/1/2"]
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=1, tags=["JOBSEQ=0000000004"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 5th loop. Nothing scheduled. But delayed jobs queue got filled with all next jobs params
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        self.assertEqual(len(manager.get_delayed_jobs()), 296)

    @patch("shub_workflow.crawl.WorkFlowManager.schedule_spider")
    def test_max_jobs_per_spider_with_known_spiders(
        self, mocked_super_schedule_spider, mocked_add_job_tags, mocked_get_jobs
    ):
        """If we limit max next params but also we know which spiders can be scheduled, we can avoid the delayed
        queue fill effect while every spider has slots available. But once one spider run out of params and
        there no running ones (loop 8th in the test), we get again the fill effect. Again, this is unavoidable if we
        don't have information about wether there are spiderA next params available."""

        spider_list = ("spiderA", "spiderB", "spiderC")

        class _ListTestManager(GeneratorCrawlManager):

            name = "test"
            max_jobs_per_spider = 1
            spider = SpiderName("")

            def set_parameters_gen(self):
                for i in range(100):
                    for spider in spider_list:
                        if spider != "spiderA" or i < 4:
                            yield {"spider": spider, "argA": i}

            def get_max_next_params(self) -> int:
                if self.get_running_spiders() == set(spider_list):
                    return 0
                return min(1, super().get_max_next_params())

        with script_args([]):
            manager = _ListTestManager()

        manager._on_start()

        # 1st loop. Schedule spiderA
        mocked_super_schedule_spider.side_effect = ["999/1/1", "999/2/1", "999/3/1"]
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 1)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=0, tags=["JOBSEQ=0000000001"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 2nd loop. Schedule spiderB
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 2)
        mocked_super_schedule_spider.assert_any_call(
            "spiderB", units=None, argA=0, tags=["JOBSEQ=0000000002"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 3rd loop. SpiderA finishes, but spiderC is next.
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/1" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 3)
        mocked_super_schedule_spider.assert_any_call(
            "spiderC", units=None, argA=0, tags=["JOBSEQ=0000000003"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 4th loop. Schedule spiderA
        mocked_super_schedule_spider.side_effect = ["999/1/2"]
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=1, tags=["JOBSEQ=0000000004"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 5th loop. Nothing scheduled, and nothing added to delayed jobs.
        manager.is_finished = lambda x: None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 4)
        self.assertEqual(len(manager.get_delayed_jobs()), 0)

        # 6th loop. SpiderA finishes and rescheduled. SpiderB and SpiderC need to go to delayed jobs.
        mocked_super_schedule_spider.side_effect = ["999/1/3"]
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/2" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 5)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=2, tags=["JOBSEQ=0000000005"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 2)
        self.assertEqual(manager.get_delayed_spiders(), {"spiderB", "spiderC"})

        # 7th loop. SpiderA finishes, reschedule one more. Two more spiderB and spiderC next params added to
        # delayed jobs
        mocked_super_schedule_spider.side_effect = ["999/1/4"]
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/3" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 6)
        mocked_super_schedule_spider.assert_any_call(
            "spiderA", units=None, argA=3, tags=["JOBSEQ=0000000006"], job_settings={}
        )
        self.assertEqual(len(manager.get_delayed_jobs()), 4)
        self.assertEqual(manager.get_delayed_spiders(), {"spiderB", "spiderC"})

        # 8th loop. SpiderA finishes, but there are no more next params for spiderA. Delayed job become filled
        # with all next params
        manager.is_finished = lambda x: Outcome("finished") if x == "999/1/4" else None
        result = next(manager._run_loops())
        self.assertTrue(result)
        self.assertEqual(mocked_super_schedule_spider.call_count, 6)
        self.assertEqual(len(manager.get_delayed_jobs()), 198)
        self.assertEqual(manager.get_delayed_spiders(), {"spiderB", "spiderC"})



import re
import logging
import uuid
from typing import List, Generator, Callable
from operator import itemgetter
from glob import iglob
from os import listdir, remove, environ, makedirs
from os.path import exists as os_exists, dirname, getctime, basename
from shutil import copyfile
from datetime import datetime, timedelta, timezone

try:
    from s3fs import S3FileSystem as OriginalS3FileSystem
    import boto3

    s3_enabled = True
except ImportError:
    s3_enabled = False


try:
    from shub_workflow.deliver import gcstorage

    gcs_enabled = True
except ImportError:
    gcs_enabled = False


S3_PATH_RE = re.compile("s3://(.+?)/(.+)")
_S3_ATTRIBUTE = "s3://"
_GS_ATTRIBUTE = "gs://"
BUFFER = 1024 * 1024


logger = logging.getLogger(__name__)


def just_log_exception(exception):
    logger.error(str(exception))
    for etype in (KeyboardInterrupt, SystemExit, ImportError):
        if isinstance(exception, etype):
            return False
    return True  # retries any other exception


if s3_enabled:

    class S3FileSystem(OriginalS3FileSystem):
        read_timeout = 120
        connect_timeout = 60


def s3_path(path, is_folder=False):
    if not path:
        return ""
    path = path.strip()
    if is_folder and not path.endswith("/"):
        path = path + "/"
    return path[len(_S3_ATTRIBUTE):]


def s3_credentials(key, secret, token, region=None):
    creds = dict(
        key=environ.get("AWS_ACCESS_KEY_ID", key), secret=environ.get("AWS_SECRET_ACCESS_KEY", secret), token=token
    )

    result = {
        **creds,
        "config_kwargs": {"retries": {"max_attempts": 20}},
    }
    region = region or environ.get("AWS_REGION")
    if region is not None:
        result.update({"client_kwargs": {"region_name": region}})
    return result


def check_s3_path(path):
    if path.startswith(_S3_ATTRIBUTE):
        if s3_enabled:
            return True
        raise ModuleNotFoundError(
            "S3 dependencies are not installed. Install shubw-workflow as shub-workflow[with-s3-tools]"
        )
    return False


def check_gcs_path(path):
    if path.startswith(_GS_ATTRIBUTE):
        if gcs_enabled:
            return True
        raise ModuleNotFoundError(
            "GCS dependencies are not installed. Install shubw-workflow as shub-workflow[with-gcs-tools]"
        )
    return False


def get_s3_bucket_keyname(s3path, aws_key, aws_secret, aws_token, **kwargs):
    region = kwargs.pop("region", None)
    region_name = kwargs.pop("region_name", region)
    botocore_session = kwargs.pop("botocore_session", None)
    profile_name = kwargs.pop("profile_name", None)

    session = boto3.Session(
        aws_access_key_id=aws_key,
        aws_secret_access_key=aws_secret,
        aws_session_token=aws_token,
        region_name=region_name,
        botocore_session=botocore_session,
        profile_name=profile_name,
    )
    s3 = session.resource("s3")
    m = S3_PATH_RE.match(s3path)
    if m:
        bucket_name, path = m.groups()
        bucket = s3.Bucket(bucket_name)
    else:
        raise ValueError(f"Bad s3 path specification: {s3path}")
    return bucket, path, kwargs.pop("op_kwargs", {})


def get_file(path, *args, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    op_kwargs = kwargs.pop("op_kwargs", {})
    region = kwargs.pop("region", None)
    if check_s3_path(path):
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        if "ACL" in op_kwargs:
            op_kwargs["acl"] = op_kwargs.pop("ACL")
        fp = fs.open(s3_path(path), *args, **op_kwargs)
        try:
            fp.name = path
        except Exception:
            pass
    else:
        fp = open(path, *args, **kwargs)

    return fp


DOWNLOAD_CHUNK_SIZE = 500 * 1024 * 1024
UPLOAD_CHUNK_SIZE = 100 * 1024 * 1024


def download_file(path, dest=None, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if dest is None:
        dest = basename(path)
    if check_s3_path(path):
        with get_file(path, "rb", aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs) as r, open(
            dest, "wb"
        ) as w:
            total = 0
            while True:
                size = w.write(r.read(DOWNLOAD_CHUNK_SIZE))
                total += size
                logger.info(f"Downloaded {total} bytes from {path}")
                if size < DOWNLOAD_CHUNK_SIZE:
                    break
    elif check_gcs_path(path):
        gcstorage.download_file(path, dest)
    else:
        raise ValueError(f"Not supported file system fpr path {path}")


def upload_file_obj(robj, dest, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if check_s3_path(dest):
        bucket, keyname, op_kwargs = get_s3_bucket_keyname(dest, aws_key, aws_secret, aws_token, **kwargs)
        bucket.upload_fileobj(robj, keyname, ExtraArgs=op_kwargs)
        logger.info(f"Uploaded file obj to {dest}.")
    else:
        raise ValueError("Not a supported cloud.")


def upload_file(path, dest, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if dest.endswith("/"):
        dest = dest + basename(path)
    if check_s3_path(dest):
        bucket, keyname, op_kwargs = get_s3_bucket_keyname(dest, aws_key, aws_secret, aws_token, **kwargs)
        bucket.upload_file(path, keyname, ExtraArgs=op_kwargs)
        logger.info(f"Uploaded {path} to {dest}.")
    elif check_gcs_path(dest):
        gcstorage.upload_file(path, dest)
    else:
        raise ValueError("Not a supported cloud.")


def get_glob(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs) -> List[str]:
    region = kwargs.pop("region", None)
    if check_s3_path(path):
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        fp = [_S3_ATTRIBUTE + p for p in fs.glob(s3_path(path))]
    else:
        fp = list(iglob(path))

    return fp


def cp_file(src_path, dest_path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if dest_path.endswith("/"):
        dest_path = dest_path + basename(src_path)
    region = kwargs.pop("region", None)
    if check_s3_path(src_path) and check_s3_path(dest_path):
        op_kwargs = kwargs.pop("op_kwargs", {})
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        fs.copy(s3_path(src_path), s3_path(dest_path), **op_kwargs)
    elif check_s3_path(src_path):
        download_file(src_path, dest_path, aws_key, aws_secret, aws_token, **kwargs)
    elif check_s3_path(dest_path):
        upload_file(src_path, dest_path, aws_key, aws_secret, aws_token, **kwargs)
    elif check_gcs_path(dest_path):
        gcstorage.upload_file(src_path, dest_path)
    else:
        dname = dirname(dest_path)
        if dname:
            makedirs(dname, exist_ok=True)
        copyfile(src_path, dest_path)


def mv_file(src_path, dest_path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if dest_path.endswith("/"):
        dest_path = dest_path + basename(src_path)
    if check_s3_path(src_path) and check_s3_path(dest_path):
        cp_file(src_path, dest_path, aws_key, aws_secret, aws_token, **kwargs)
        rm_file(src_path, aws_key, aws_secret, aws_token, **kwargs)
    elif check_s3_path(src_path):
        download_file(src_path, dest_path, aws_key, aws_secret, aws_token, **kwargs)
        rm_file(src_path, aws_key, aws_secret, aws_token, **kwargs)
    elif check_s3_path(dest_path):
        upload_file(src_path, dest_path, aws_key, aws_secret, aws_token, **kwargs)
        rm_file(src_path)
    elif check_gcs_path(src_path) and check_gcs_path(dest_path):
        gcstorage.mv_file(src_path, dest_path)
    elif check_gcs_path(src_path):
        gcstorage.download_file(src_path, dest_path)
        gcstorage.rm_file(src_path)
    elif check_gcs_path(dest_path):
        gcstorage.upload_file(src_path, dest_path)
        rm_file(src_path)
    else:
        dname = dirname(dest_path)
        if dname:
            makedirs(dname, exist_ok=True)
        cp_file(src_path, dest_path)
        rm_file(src_path)


def rm_file(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    region = kwargs.pop("region", None)
    if check_s3_path(path):
        op_kwargs = kwargs.pop("op_kwargs", {})
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        fs.rm(s3_path(path), **op_kwargs)
    elif check_gcs_path(path):
        gcstorage.rm_file(path)
    else:
        remove(path)


def list_path(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs) -> Generator[str, None, None]:
    """
    More efficient boto3 based path listing, that accepts prefix
    """
    if check_s3_path(path):
        bucket, path, op_kwargs = get_s3_bucket_keyname(path, aws_key, aws_secret, aws_token, **kwargs)
        for result in bucket.objects.filter(Prefix=path):
            yield f"s3://{bucket.name}/{result.key}"
    elif check_gcs_path(path):
        yield from gcstorage.list_path(path)
    else:
        listing = []
        if path.strip().endswith("/"):
            path = path[:-1]
        try:
            listing = [f"{path}/{name}" for name in listdir(path)]
        except FileNotFoundError:
            pass
        yield from listing


def list_folder(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs) -> List[str]:
    if check_s3_path(path):
        region = kwargs.pop("region", None)
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)

        try:
            path = s3_path(path, is_folder=True)
            listing = [f"s3://{name}" for name in fs.ls(path) if name != path]
        except FileNotFoundError:
            listing = []
    elif check_gcs_path(path):
        listing = list(gcstorage.list_folder(path))
    else:
        if path.strip().endswith("/"):
            path = path[:-1]
        try:
            listing = [f"{path}/{name}" for name in listdir(path)]
        except FileNotFoundError:
            listing = []

    return listing


def list_folder_in_ts_order(
    input_folder: str, aws_key=None, aws_secret=None, aws_token=None, **kwargs
) -> Generator[str, None, None]:
    results = []
    for input_file in list_folder(input_folder, aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs):
        if check_s3_path(input_folder):
            with get_file(input_file, "rb", aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs) as f:
                results.append((input_file, f.info()["LastModified"]))
        else:
            results.append((input_file, getctime(input_file)))
    for n, _ in sorted(results, key=itemgetter(1)):
        yield n


def list_folder_files_recursive(
    folder: str, aws_key=None, aws_secret=None, aws_token=None, **kwargs
) -> Generator[str, None, None]:
    for f in list_folder(folder, aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs):
        if f == folder:
            yield f
        else:
            yield from list_folder_files_recursive(
                f, aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs
            )


def s3_folder_size(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    region = kwargs.pop("region", None)
    if check_s3_path(path):
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        return sum(fs.du(s3_path(path, is_folder=True), deep=True).values())


def exists(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    region = kwargs.pop("region", None)
    if check_s3_path(path):
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        return fs.exists(path)
    return os_exists(path)


def empty_folder(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs) -> List[str]:
    region = kwargs.pop("region", None)
    removed_files = []
    if check_s3_path(path):
        fs = S3FileSystem(**s3_credentials(aws_key, aws_secret, aws_token, region), **kwargs)
        for s3file in list_folder(path, aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token):
            try:
                fs.rm(s3file)
                removed_files.append(s3file)
            except FileNotFoundError:
                pass
    else:
        if path.strip().endswith("/"):
            path = path[:-1]
        for fsfile in listdir(path):
            remove(f"{path}/{fsfile}")
            removed_files.append(f"{path}/{fsfile}")

    return removed_files


def touch(path, aws_key=None, aws_secret=None, aws_token=None, **kwargs):
    if not exists(path):
        if not check_s3_path(path):
            dname = dirname(path)
            makedirs(dname, exist_ok=True)
        with get_file(path, "w", aws_key=aws_key, aws_secret=aws_secret, aws_token=aws_token, **kwargs) as f:
            f.write("")
    else:
        raise ValueError(f"File {path} already exists")


class S3SessionFactory:

    """
    Generate s3 temporal session credentials in cases where that is required
    """

    def __init__(self, aws_key, aws_secret, aws_role, aws_external_id, expiration_margin_minutes=10):
        self.aws_role = aws_role
        self.aws_key = aws_key
        self.aws_secret = aws_secret
        self.aws_external_id = aws_external_id
        self.expiration_margin_minutes = expiration_margin_minutes
        self.credentials_expiration = None
        self.sts_client = boto3.client("sts", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
        self.full_credentials = {}
        self.s3_client = None

    def get_credentials(self):
        """
        Get a new set of credentials if the existing ones are close to expire,
        update the S3 client.
        """
        if self.credentials_expiration is None or datetime.now(timezone.utc) > self.credentials_expiration - timedelta(
            minutes=self.expiration_margin_minutes
        ):
            session_name = str(uuid.uuid4())
            self.full_credentials = self.sts_client.assume_role(
                RoleArn=self.aws_role,
                RoleSessionName=session_name,
                ExternalId=self.aws_external_id,
            )["Credentials"]
            self.credentials_expiration = self.full_credentials["Expiration"]
            self.s3_client = boto3.session.Session(
                aws_access_key_id=self.full_credentials["AccessKeyId"],
                aws_secret_access_key=self.full_credentials["SecretAccessKey"],
                aws_session_token=self.full_credentials["SessionToken"],
            ).client("s3")
        return {
            "aws_key": self.full_credentials["AccessKeyId"],
            "aws_secret": self.full_credentials["SecretAccessKey"],
            "aws_token": self.full_credentials["SessionToken"],
        }


class S3Helper:

    cp_file: Callable
    download_file: Callable
    empty_folder: Callable
    exists: Callable
    get_file: Callable
    get_glob: Callable
    list_folder: Callable
    list_folder_files_recursive: Callable
    list_folder_in_ts_order: Callable
    list_path: Callable
    mv_file: Callable
    rm_file: Callable
    s3_folder_size: Callable
    touch: Callable
    upload_file: Callable
    upload_file_obj: Callable

    def __init__(
        self,
        aws_key,
        aws_secret,
        aws_role=None,
        aws_external_id=None,
        expiration_margin_minutes=10,
        op_kwargs_by_method_name=None,
    ):
        self.s3_session_factory = None
        self.credentials = {"aws_key": aws_key, "aws_secret": aws_secret}
        if aws_role is not None and aws_external_id is not None:
            self.s3_session_factory = S3SessionFactory(
                aws_key, aws_secret, aws_role, aws_external_id, expiration_margin_minutes
            )

        op_kwargs_by_method_name = op_kwargs_by_method_name or {}
        for method_name, _type in S3Helper.__annotations__.items():
            if not hasattr(self, method_name) and _type is Callable:
                method = globals()[method_name]
                self._wrap_method(method, **op_kwargs_by_method_name.get(method_name, {}))

    def _wrap_method(self, method, **op_kwargs):
        def _method(*args, **kwargs):
            if self.s3_session_factory is not None:
                self.credentials = self.s3_session_factory.get_credentials()
            if op_kwargs:
                kwargs.update({"op_kwargs": op_kwargs})
            kwargs.update(self.credentials)
            return method(*args, **kwargs)

        setattr(self, method.__name__, _method)


class FSHelper(S3Helper):
    def __init__(self, aws_key=None, aws_secret=None, **kwargs):
        super().__init__(aws_key, aws_secret, **kwargs)



"""
Implements common methods for any workflow manager
"""
import abc
import time
import logging
from uuid import uuid4
from argparse import Namespace
from collections import defaultdict
from typing import Optional, Generator, Protocol, List, Union, Dict, Tuple

from .script import BaseLoopScript, JobKey, JobDict, Outcome, BaseLoopScriptProtocol


logger = logging.getLogger(__name__)


class WorkFlowManagerProtocol(BaseLoopScriptProtocol, Protocol):
    @abc.abstractmethod
    def get_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
        ...

    @abc.abstractmethod
    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
        ...

    @property
    @abc.abstractmethod
    def max_running_jobs(self) -> int:
        ...


class CachedFinishedJobsMixin(WorkFlowManagerProtocol):
    def __init__(self):
        super().__init__()
        self.__finished_cache: Dict[JobKey, Outcome] = {}
        self.__update_finished_cache_called: Dict[int, bool] = defaultdict(bool)

    def update_finished_cache(self, project_id: int):
        if not self.__update_finished_cache_called[project_id]:
            logger.info("Initiating finished cache update.")
            for job in self.get_owned_jobs(project_id, state=["finished"], meta=["close_reason"]):
                if job["key"] in self.__finished_cache:
                    break
                self.__finished_cache[job["key"]] = Outcome(job["close_reason"])
            logger.info(f"Finished jobs cache length: {len(self.__finished_cache)}")
        self.__update_finished_cache_called[project_id] = True

    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
        kwargs.setdefault("meta", []).append("close_reason")
        # only use for updating finished cache if no filter/limit imposed.
        update_finished_cache = not set(kwargs.keys()).intersection(["count", "has_tag", "lacks_tag"])
        finished_cache = []
        for job in super().get_finished_owned_jobs(project_id, **kwargs):  # type: ignore
            if update_finished_cache:
                finished_cache.append((job["key"], Outcome(job["close_reason"])))
            yield job
        logger.info(f"Preread {len(finished_cache)} finished jobs.")
        while finished_cache:
            key, close_reason = finished_cache.pop()
            self.__finished_cache[key] = close_reason
        logger.info(f"Finished jobs cache length: {len(self.__finished_cache)}")

    def is_finished(self, jobkey: JobKey) -> Optional[Outcome]:
        project_id = int(jobkey.split("/", 1)[0])
        self.update_finished_cache(project_id)
        return self.__finished_cache.get(jobkey)

    def base_loop_tasks(self):
        super().base_loop_tasks()
        for project_id in self.__update_finished_cache_called.keys():
            self.__update_finished_cache_called[project_id] = False


class WorkFlowManager(BaseLoopScript, WorkFlowManagerProtocol):

    # --max-running-job command line option overrides it
    default_max_jobs: int = 1000

    flow_id_required = True

    base_failed_outcomes: Tuple[str, ...] = (
        "failed",
        "killed by oom",
        "cancelled",
        "cancel_timeout",
        "memusage_exceeded",
        "diskusage_exceeded",
        "cancelled (stalled)",
    )

    def __init__(self):
        self.failed_outcomes = list(self.base_failed_outcomes)
        self.is_resumed = False
        self.__autogenerated_flow_id = False
        super().__init__()

    def set_flow_id_name(self, args):
        super().set_flow_id_name(args)
        if not self.name:
            self.argparser.error("Manager name not set.")

    def get_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
        assert self.flow_id, "This job doesn't have a flow id."
        assert self.name, "This job doesn't have a name."
        assert "has_tag" not in kwargs, "Filtering by flow id requires no extra has_tag."
        assert "state" in kwargs, "'state' parameter must be provided."
        kwargs["has_tag"] = [f"FLOW_ID={self.flow_id}"]
        parent_tag = f"PARENT_NAME={self.name}"
        meta = kwargs.get("meta") or []
        if "tags" not in meta:
            meta.append("tags")
        kwargs["meta"] = meta
        for job in self.get_jobs(project_id, **kwargs):
            if parent_tag in job["tags"]:
                yield job

    def generate_flow_id(self) -> str:
        self.__autogenerated_flow_id = True
        return str(uuid4())

    @property
    def max_running_jobs(self) -> int:
        return self.args.max_running_jobs

    def add_argparser_options(self):
        super().add_argparser_options()
        if not self.name:
            self.argparser.add_argument("name", help="Script name.")
        self.argparser.add_argument(
            "--max-running-jobs",
            type=int,
            default=self.default_max_jobs,
            help="If given, don't allow more than the given jobs running at once.\
                                    Default: %(default)s",
        )

    def parse_args(self) -> Namespace:
        args = super().parse_args()
        if not self.name:
            self.name = args.name
        return args

    def wait_for(
        self,
        jobs_keys: Union[JobKey, List[JobKey]],
        interval: int = 60,
        timeout: float = float("inf"),
        heartbeat: Optional[int] = None,
    ):
        """Waits until all given jobs are not running anymore or until the
        timeout is reached, if a heartbeat is given it'll log an entry every
        heartbeat seconds (considering the interval), otherwise it'll log an
        entry every interval seconds.
        """
        if isinstance(jobs_keys, str):
            jobs_keys = [jobs_keys]
        if heartbeat is None or heartbeat < interval:
            heartbeat = interval
        still_running = dict((key, True) for key in jobs_keys)
        time_waited, next_heartbeat = 0, heartbeat
        while any(still_running.values()) and time_waited < timeout:
            time.sleep(interval)
            time_waited += interval
            for key in (k for k, v in still_running.items() if v):
                if self.is_running(key):
                    if time_waited >= next_heartbeat:
                        next_heartbeat += heartbeat
                        logger.info(f"{key} still running")
                    break
                still_running[key] = False

    def _check_resume_workflow(self):
        for job in self.get_jobs(state=["finished"], meta=["tags"], has_tag=[f"NAME={self.name}"]):
            if self.get_keyvalue_job_tag("FLOW_ID", job["tags"]) == self.flow_id:
                inherited_tags = []
                for tag in job["tags"]:
                    if len(tag.split("=")) == 2:
                        inherited_tags.append(tag)
                self.add_job_tags(tags=inherited_tags)
                self.is_resumed = True
                break

    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
        kwargs.pop("state", None)
        for job in self.get_owned_jobs(project_id, state=["finished"], **kwargs):
            yield job

    def resume_workflow(self):
        rcount = 0
        for job in self.get_owned_jobs(state=["running", "pending"], meta=["spider_args", "job_cmd", "tags", "spider"]):
            self.resume_running_job_hook(job)
            logger.info(f"Found running job {job['key']}")
            rcount += 1
        if rcount > 0:
            logger.info(f"Found a total of {rcount} running children jobs.")

        fcount = 0
        logger.info("Searching finished children jobs...")
        for job in self.get_finished_owned_jobs(meta=["spider_args", "job_cmd", "tags", "spider", "close_reason"]):
            self.resume_finished_job_hook(job)
            fcount += 1
        if fcount > 0:
            logger.info(f"Found a total of {fcount} finished children jobs.")

    def resume_running_job_hook(self, job: JobDict):
        pass

    def resume_finished_job_hook(self, job: JobDict):
        pass

    def _on_start(self):
        if not self.__autogenerated_flow_id:
            self._check_resume_workflow()
        if self.is_resumed:
            self.resume_workflow()
        super()._on_start()



"""
Utility for cloning ScrapyCloud jobs
Features tagging of cloned from/to jobs (both source and destination) and avoids to clone source jobs already cloned.
By default cloned jobs are scheduled in the same project as source job. If --project-id is given, target project
is overriden.
"""
import logging
from typing import Optional, List

from scrapinghub.client.jobs import Job
from scrapinghub import DuplicateJobError

from shub_workflow.script import BaseScript, JobKey
from shub_workflow.utils import dash_retry_decorator


_LOG = logging.getLogger(__name__)


def _transform_cmd(job_cmd):
    if isinstance(job_cmd, list):
        return " ".join(["'%s'" % cmd for cmd in job_cmd[1:]])

    return job_cmd


_COPIED_FROM_META = {
    "job_cmd": ("cmd_args", _transform_cmd),
    "units": (None, None),
    "spider_args": ("job_args", None),
    "tags": ("add_tag", None),
    "job_settings": (None, None),
}


class BaseClonner(BaseScript):

    MAX_CLONES = 10

    def is_cloned(self, jobkey: JobKey):
        for tag in self.get_job_tags(jobkey):
            if tag.startswith("ClonedTo="):
                _LOG.warning(f"Job {jobkey} already cloned. Skipped.")
                return True
        return False

    def job_params_hook(self, job_params):
        pass

    def clone_job(
        self, job_key: JobKey, units: Optional[int] = None, extra_tags: Optional[List[str]] = None
    ) -> Optional[Job]:
        extra_tags = extra_tags or []
        job = self.get_job(job_key)

        spider = self._get_metadata_key(job.metadata, "spider")

        job_params = dict()
        for key, (target_key, _) in _COPIED_FROM_META.items():

            if target_key is None:
                target_key = key

            job_params[target_key] = self._get_metadata_key(job.metadata, key)

        clone_number = 0
        add_tag = job_params.setdefault("add_tag", [])

        add_tag = list(filter(lambda x: not x.startswith("ClonedFrom="), add_tag))
        add_tag.append(f"ClonedFrom={job_key}")

        for tag in add_tag:
            if tag.startswith("CloneNumber="):
                clone_number = int(tag.replace("CloneNumber=", ""))
                break

        clone_number += 1

        if clone_number >= self.MAX_CLONES:
            _LOG.warning(f"Already reached max clones allowed for job {job_key}.")
            return None

        add_tag = list(filter(lambda x: not x.startswith("CloneNumber="), add_tag))
        add_tag.append(f"CloneNumber={clone_number}")

        add_tag.extend(extra_tags)
        job_params["add_tag"] = add_tag
        if units is not None:
            job_params["units"] = units

        self.job_params_hook(job_params)

        for key, (target_key, transform) in _COPIED_FROM_META.items():

            target_key = target_key or key

            if transform is None:

                def transform(x):
                    return x

            job_params[target_key] = transform(job_params[target_key])

        project_id, _, _ = job_key.split("/")
        project = self.get_project(self.project_id or project_id)
        new_job = self.schedule_generic(project, spider, **job_params)
        if new_job is not None:
            _LOG.info("Cloned %s to %s", job_key, new_job.key)
            jobtags = self._get_metadata_key(job.metadata, "tags")
            jobtags.append(f"ClonedTo={new_job.key}")
            self._update_metadata(job.metadata, {"tags": jobtags})

        return new_job

    @dash_retry_decorator
    def schedule_generic(self, project, spider, **job_params) -> Optional[Job]:
        try:
            return project.jobs.run(spider, **job_params)
        except DuplicateJobError as e:
            _LOG.error(str(e))
        return None


class CloneJobScript(BaseClonner):

    flow_id_required = False

    @property
    def description(self):
        return __doc__

    def parse_project_id(self, args):
        project_id = super().parse_project_id(args)
        if project_id:
            return project_id
        if args.key:
            return args.key[0].split("/")[0]

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument(
            "key",
            type=str,
            nargs="+",
            default=[],
            help="Target job key. Can be given multiple times. All must be in same project.",
        )
        self.argparser.add_argument("--units", help="Set number of units. Default is the same as cloned job.", type=int)

    def run(self):
        keys = list(filter(lambda x: not self.is_cloned(x), self.args.key))
        for job_key in keys:
            try:
                self.clone_job(job_key, self.args.units, self.args.children_tag)
            except Exception as e:
                _LOG.error("Could not restart job %s: %s", job_key, e)


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s %(name)s [%(levelname)s]: %(message)s",
        level=logging.DEBUG,
    )
    script = CloneJobScript()
    script.run()
